import logging
import shutil
from typing import Optional, List
from enum import Enum, auto

from .plugins import LanguagePluginManager, LanguageEnvironmentPluginManager, VirtualEnvironment
from pydotfiles.common import OS, PackageManager
from pydotfiles.utils import run_command, run_command_with_communication


logger = logging.getLogger(__name__)


class LanguageManager(Enum):
    """
    Represents a language manager like
    pyenv, rbenv, goenv, jenv, etc.
    """
    PYENV = auto()
    RBENV = auto()
    GOENV = auto()
    JENV = auto()

    @staticmethod
    def from_string(label):
        if label is None:
            raise KeyError("Enum: No language manager passed in")

        return LanguageManager[label.upper()]


class LanguageEnvironmentManager:

    def __init__(self, language_manager: LanguageManager, sudo_password="", language_plugin_managers: Optional[List[LanguageEnvironmentPluginManager]] = None):

        if language_plugin_managers is None:
            language_plugin_managers = []

        self.language_manager = language_manager
        self.sudo_password = sudo_password
        self.language_plugin_managers = language_plugin_managers

    @property
    def is_installed(self):
        if self.language_manager == LanguageManager.PYENV:
            return shutil.which("pyenv") is not None
        elif self.language_manager == LanguageManager.JENV:
            return shutil.which("jenv") is not None
        else:
            raise NotImplementedError("Language Environment Manager: Sorry, the requested language manager is not currently supported (feel free to open an issue at https://github.com/JasonYao/pydotfiles/issues/new)")

    def install(self):
        if self.is_installed:
            logger.info(f"Language Environment Manager: Already installed language environment manager [env_manager={self.language_manager}]")
            return

        install_language_environment_manager(self.language_manager)

    def install_plugins(self):
        for language_plugin in self.language_plugin_managers:
            language_plugin.install()
            language_plugin.install_virtual_environments()

    def install_language_version(self, version):
        install_language(self.language_manager, version, self.sudo_password)

    def uninstall(self):
        if not self.is_installed:
            logger.info(f"Language Environment Manager: Already uninstalled language environment manager [env_manager={self.language_manager}]")
            return

        uninstall_language_environment_manager(self.language_manager)

    def uninstall_plugins(self):
        for language_plugin in self.language_plugin_managers:
            language_plugin.uninstall_virtual_environments()
            language_plugin.uninstall()

    def uninstall_language_version(self, version):
        uninstall_language(self.language_manager, version, self.sudo_password)


class DevelopmentEnvironment:
    """
    Class representing a dev environment, such as
    python, ruby, or java
    """

    def __init__(self, language: str, versions: List[str], language_environment_manager: LanguageEnvironmentManager):
        self.language = language
        self.versions = versions
        self.language_environment_manager = language_environment_manager

    def install(self):
        logger.info(f"Development Environment: Installing [lang={self.language}, versions={self.versions}, env_manager={self.language_environment_manager}")
        self.language_environment_manager.install()

        # Installs the base version
        for version in self.versions:
            self.language_environment_manager.install_language_version(version)

        # Installs any plugins
        self.language_environment_manager.install_plugins()

        logger.info(f"Development Environment: Successfully installed [lang={self.language}, versions={self.versions}, env_manager={self.language_environment_manager}")

    def uninstall(self):
        logger.info(f"Development Environment: Uninstalling [lang={self.language}, versions={self.versions}, env_manager={self.language_environment_manager}")

        # Uninstalls any plugins
        self.language_environment_manager.uninstall_plugins()

        # Uninstalls the base version
        for version in self.versions:
            self.language_environment_manager.uninstall_language_version(version)

        self.language_environment_manager.uninstall()
        logger.info(f"Development Environment: Successfully uninstalled [lang={self.language}, versions={self.versions}, env_manager={self.language_environment_manager}")


"""
Switching-layer functions
"""


def install_language_environment_manager(env_manager):
    host_system = OS.get_host_system()
    package_manager = OS.get_package_manager(host_system)

    if package_manager == PackageManager.BREW:
        install_language_environment_manager_via_homebrew(env_manager)
    else:
        raise NotImplementedError(f"Language Environment Manager: Sorry, the requested package manager is not currently supported [package_manager={package_manager}]")


def install_language(env_manager, language_version, sudo_password=""):
    if env_manager == LanguageManager.PYENV:
        check_command = f"pyenv versions"
        check_response = run_command(check_command)

        if language_version in check_response:
            logger.info(f"Language Environment Manager: Already installed language version [language=python, version={language_version}]")
            return

        logger.info(f"Language Environment Manager: Installing language version (this might take a bit, ~2 min) [language=python, version={language_version}]")

        command = f"pyenv install {language_version}"
        run_command(command)

        logger.info(f"Language Environment Manager: Successfully installed language version [language=python, version={language_version}]")
    elif env_manager == LanguageManager.JENV:
        host_system = OS.get_host_system()
        package_manager = OS.get_package_manager(host_system)

        if package_manager == PackageManager.BREW:
            install_java_version_via_homebrew(language_version, sudo_password)
        else:
            raise NotImplementedError(f"Language Environment Manager: Sorry, the requested package manager is not currently supported [package_manager={package_manager}]")
    else:
        raise NotImplementedError("Language Environment Manager: Sorry, the requested language manager is not currently supported (feel free to open an issue at https://github.com/JasonYao/pydotfiles/issues/new)")


def uninstall_language_environment_manager(env_manager):
    host_system = OS.get_host_system()
    package_manager = OS.get_package_manager(host_system)

    if package_manager == PackageManager.BREW:
        uninstall_language_environment_manager_via_homebrew(env_manager)
    else:
        raise NotImplementedError(f"Language Environment Manager: Sorry, the requested package manager is not currently supported [package_manager={package_manager}]")


def uninstall_language(env_manager, language_version, sudo_password=""):
    if env_manager == LanguageManager.PYENV:
        check_command = f"pyenv versions"
        check_response = run_command(check_command)

        if language_version not in check_response:
            logger.info(f"Language Environment Manager: Already uninstalled language version [language=python, version={language_version}]")
            return

        logger.info(f"Language Environment Manager: Uninstalling language version [language=python, version={language_version}]")

        command = f"pyenv uninstall {language_version}"
        run_command_with_communication(command, "y\n")  # The 'y\n' is to confirm that we want to uninstall the given python version

        logger.info(f"Language Environment Manager: Successfully uninstalled language version [language=python, version={language_version}]")
    elif env_manager == LanguageManager.JENV:
        host_system = OS.get_host_system()
        package_manager = OS.get_package_manager(host_system)

        if package_manager == PackageManager.BREW:
            uninstall_java_version_via_homebrew(language_version, sudo_password)
        else:
            raise NotImplementedError(f"Language Environment Manager: Sorry, the requested package manager is not currently supported [package_manager={package_manager}]")
    else:
        raise NotImplementedError("Language Environment Manager: Sorry, the requested language manager is not currently supported (feel free to open an issue at https://github.com/JasonYao/pydotfiles/issues/new)")


"""
Homebrew functions
"""

LATEST_CURRENT_JAVA_VERSION_IN_HOMEBREW = "11"


def install_language_environment_manager_via_homebrew(env_manager):
    logger.info(f"Language Environment Manager: Installing environment manager [env_manager={env_manager}]")
    if env_manager == LanguageManager.PYENV:
        command = "brew install pyenv"
        run_command(command)
    elif env_manager == LanguageManager.JENV:
        command = "brew install jenv"
        run_command(command)
    else:
        raise NotImplementedError("Language Environment Manager: Sorry, the requested language manager is not currently supported (feel free to open an issue at https://github.com/JasonYao/pydotfiles/issues/new)")

    logger.info(f"Language Environment Manager: Successfully installed environment manager [env_manager={env_manager}]")


def uninstall_language_environment_manager_via_homebrew(env_manager):
    if env_manager == LanguageManager.PYENV:
        command = "brew uninstall pyenv"
        run_command(command)
    elif env_manager == LanguageManager.JENV:
        command = "brew uninstall jenv"
        run_command(command)
    else:
        raise NotImplementedError("Language Environment Manager: Sorry, the requested language manager is not currently supported (feel free to open an issue at https://github.com/JasonYao/pydotfiles/issues/new)")


def install_java_version_via_homebrew(language_version, sudo_password=""):
    # Checks to make sure that we haven't already installed this java version
    check_command = f"brew cask list"
    check_response = run_command(check_command)

    if (language_version == "latest" or language_version == LATEST_CURRENT_JAVA_VERSION_IN_HOMEBREW) and "java " in check_response:
        logger.info(f"Language Environment Manager: Already installed language version [language=java, version={language_version}]")
        return
    elif language_version == "8" and "java8" in check_response:
        logger.info(f"Language Environment Manager: Already installed language version [language=java, version={language_version}]")
        return

    # Actually installs java
    logger.info(f"Language Environment Manager: Installing language version (takes a bit, ~2 min) [language=java, version={language_version}]")

    if language_version != "8" and language_version != LATEST_CURRENT_JAVA_VERSION_IN_HOMEBREW and language_version != "latest":
        raise ValueError(f"Language Environment Manager: Due to a limitation in Homebrew, we are unable to install any Java versions besides Java 8 and the latest Java (currently version {LATEST_CURRENT_JAVA_VERSION_IN_HOMEBREW})")

    if language_version == LATEST_CURRENT_JAVA_VERSION_IN_HOMEBREW or language_version == "latest":
        command = f"brew cask install java"
        run_command_with_communication(command, sudo_password)
        return

    # Getting java 8 requires getting versioned casks (note that there isn't a java9, java10, or java11 versioned cask)
    command = f"brew tap homebrew/cask-versions"
    run_command(command)

    command = f"brew cask install java8"
    run_command_with_communication(command, sudo_password)

    logger.info(f"Language Environment Manager: Successfully installed language version [language=java, version={language_version}]")


def uninstall_java_version_via_homebrew(language_version, sudo_password=""):
    logger.info(f"Language Environment Manager: Uninstalling language version [language=java, version={language_version}]")

    if language_version != "8" and language_version != LATEST_CURRENT_JAVA_VERSION_IN_HOMEBREW and language_version != "latest":
        raise ValueError(f"Language Environment Manager: Due to a limitation in Homebrew, we are unable to uninstall any Java versions besides Java 8 and the latest Java (currently version {LATEST_CURRENT_JAVA_VERSION_IN_HOMEBREW})")

    if language_version == LATEST_CURRENT_JAVA_VERSION_IN_HOMEBREW or language_version == "latest":
        command = f"brew cask uninstall java"
        run_command_with_communication(command, sudo_password)
        return

    # Getting java 8 requires getting versioned casks (note that there isn't a java9, java10, or java11 versioned cask)
    command = f"brew tap homebrew/cask-versions"
    run_command(command)

    command = f"brew cask uninstall java8"
    run_command_with_communication(command, sudo_password)

    logger.info(f"Language Environment Manager: Successfully uninstalled language version [language=java, version={language_version}]")



# General imports
from typing import Dict, Any
from pathlib import Path
from logging import getLogger

# Project imports
from pydotfiles.v4.common import get_configuration_version
from pydotfiles.v4.common import Configuration
from pydotfiles.v4.common.alpha import AlphaCore, AlphaDefaultSettings, DefaultSettings
from pydotfiles.v4.validator.alpha import AlphaValidator

logger = getLogger(__name__)

##
# Public methods
##


def validate(raw_data: Dict) -> list[Configuration]:

    validated_data = {file_path: file_data for file_path, file_data in raw_data.items() if __validate_configuration(file_path, file_data)}
    mapped_data = [Configuration(file_path, __map_data(file_data)) for file_path, file_data in validated_data.items()]
    joined_data = __join_dev_data(mapped_data)
    # We want to join our default data files in the OperatingSystem object with the AlphaDevEnvironment object
    # for mapped_datum in mapped_data:
    #     print(f"--------------------------\n")
    #     print(f"The data is: {mapped_datum}")
    #     print(f"The raw data is: {dumps(validated_data.get(mapped_datum.file_path), indent=4, sort_keys=True)}")
    #     print(f"--------------------------\n")
    logger.debug("All valid config files loaded")
    return joined_data

##
# Helper methods
##


def __validate_configuration(file_path: Path, file_data: Dict) -> bool:
    validator = get_validator(file_data)
    return validator.validate(file_path, file_data)


def get_validator(data: dict):
    """
    We only support alpha for now
    """
    return AlphaValidator()


def __map_data(validated_data: Dict) -> Any:
    configuration_version_type = get_configuration_version(validated_data.get("version"), validated_data.get("schema"))
    return configuration_version_type.from_dict(validated_data)


def __join_dev_data(data: list[Configuration]) -> list[Configuration]:
    os_map: Dict[Path, AlphaCore] = {}
    dev_env_map: Dict[Path, AlphaDefaultSettings] = {}

    for datum in data:
        if isinstance(datum.data, AlphaCore) and datum.data.os is not None and len(datum.data.os.default_settings_files) > 0:
            os_map[datum.file_path] = datum.data
        elif isinstance(datum.data, AlphaDefaultSettings):
            dev_env_map[datum.file_path] = datum.data

    final_join_dev_data: list[Configuration] = []
    for datum in data:
        if dev_env_map.get(datum.file_path) is not None:
            # Skips over this config since it's taken care of in os
            continue
        if os_map.get(datum.file_path) is None:
            final_join_dev_data.append(datum)
        else:
            # Now needs to reform the AlphaCore object with the corresponding AlphaDefaultSettings
            final_join_dev_data.append(__join_dev_datum(datum, dev_env_map))
    return final_join_dev_data


def __join_dev_datum(datum: Configuration, dev_env_map: Dict[Path, AlphaDefaultSettings]) -> Configuration:
    data: AlphaCore = datum.data

    raw_additional_data: list[AlphaDefaultSettings] = []
    for default_settings_file_path in data.os.default_settings_files:
        if dev_env_map.get(datum.file_path.parent.joinpath(default_settings_file_path)) is None:
            logger.error(f"Unable to map {data.os.default_settings_files} to a loaded-in default settings object")
        else:
            raw_additional_data.append(dev_env_map.get(datum.file_path.parent.joinpath(default_settings_file_path)))

    additional_data: list[DefaultSettings] = []
    for raw_additional_datum in raw_additional_data:
        for default_settings in raw_additional_datum.default_settings:
            additional_data.append(default_settings)

    return Configuration(file_path=datum.file_path, data=data.join_with_default_settings_data(additional_data))



from .version import VERSION_NUMBER

name = "pydotfiles"

__all__ = ['api', 'models', 'utils']

# Enables single-sourcing of the version number: https://packaging.python.org/guides/single-sourcing-package-version/
__version__ = VERSION_NUMBER



# General imports
import os
import os.path
import shutil
import subprocess

# Project imports
from .general import hash_file


"""
I/O-related public access helper methods
"""


def mv_file(origin, destination, use_sudo=False, sudo_password=""):
    # Fast fail if invalid origin/destinations passed in
    if origin is None or destination is None:
        raise RuntimeError(f"File Moving: Invalid file paths passed in [origin={origin}, destination={destination}]")

    # Fast return if there is no need for the operation
    if is_moved(origin, destination):
        return

    # Fail fast if a destination file already exists
    if os.path.isfile(destination) or os.path.islink(destination):
        raise RuntimeError(f"File Moving: Destination file already exists [origin={origin}, destination={destination}]")

    if use_sudo:
        command = f"mv {origin} {destination}"
        process = subprocess.Popen(['sudo', '-S'] + command.split(), stdin=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)

        try:
            stdout, stderr = process.communicate(sudo_password + '\n', timeout=3)

            if "File exists" in stderr:
                raise FileExistsError(f"The file {destination} already exists")

            if process.returncode != 0:
                raise RuntimeError(stderr)
        except subprocess.TimeoutExpired:
            process.kill()
            raise
    else:
        shutil.move(origin, destination)


def rm_file(file, use_sudo=False, sudo_password=""):
    # Fast fail if invalid file is passed in
    if file is None:
        raise RuntimeError(f"File Removing: Invalid file passed in [file={file}]")

    # Fast return if there is no need for the operation
    if not os.path.isfile(file) and not is_broken_link(file):
        return

    if use_sudo:
        command = f"rm {file}"
        process = subprocess.Popen(['sudo', '-S'] + command.split(), stdin=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)

        try:
            stdout, stderr = process.communicate(sudo_password + '\n', timeout=3)

            if process.returncode != 0:
                raise RuntimeError(stderr)
        except subprocess.TimeoutExpired:
            process.kill()
            raise
    else:
        os.unlink(file)


def copy_file(origin, destination, use_sudo=False, sudo_password=""):
    # Fast fail if invalid origin/destinations passed in
    if origin is None or destination is None:
        raise RuntimeError(f"File Copying: Invalid symlinking file paths passed in [origin={origin}, destination={destination}]")

    # Fast return if there is no need for the operation
    if is_copied(origin, destination):
        return

    # Fast fail if the file already exists
    if os.path.isfile(destination) or os.path.islink(destination):
        raise RuntimeError(f"File Copying: Destination file already exists [origin={origin}, destination={destination}]")

    if use_sudo:
        command = f"cp {origin} {destination}"
        process = subprocess.Popen(['sudo', '-S'] + command.split(), stdin=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)

        try:
            stdout, stderr = process.communicate(sudo_password + '\n', timeout=3)

            if "File exists" in stderr:
                raise FileExistsError(f"The file {destination} already exists")

            if process.returncode != 0:
                raise RuntimeError(stderr)
        except subprocess.TimeoutExpired:
            process.kill()
            raise
    else:
        shutil.copy2(origin, destination)


def symlink_file(origin, destination, use_sudo=False, sudo_password=""):
    # Fast fail if invalid origin/destinations passed in
    if origin is None or destination is None:
        raise RuntimeError(f"File Symlinking: Invalid symlinking file paths passed in [origin={origin}, destination={destination}]")

    # Fast return if there is no need for the operation
    if is_linked(origin, destination):
        return

    # Fast fail if the file already exists
    if os.path.isfile(destination) or os.path.islink(destination):
        raise RuntimeError(f"File Symlinking: Destination file already exists [origin={origin}, destination={destination}]")

    if use_sudo:
        command = f"ln -s {origin} {destination}"
        process = subprocess.Popen(['sudo', '-S'] + command.split(), stdin=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)

        try:
            stdout, stderr = process.communicate(sudo_password + '\n', timeout=3)

            if "File exists" in stderr:
                raise FileExistsError(f"The file {destination} already exists")

            if process.returncode != 0:
                raise RuntimeError(stderr)
        except subprocess.TimeoutExpired:
            process.kill()
            raise
    else:
        os.symlink(origin, destination)


def unsymlink_file(file, use_sudo=False, sudo_password=""):
    # Fast fail if invalid file name or type passed in
    if file is None or not os.path.islink(file):
        raise RuntimeError(f"File Unsymlinking: File does not exist or is a symlink [file={file}]")

    if use_sudo:
        command = f"unlink {file}"
        process = subprocess.Popen(['sudo', '-S'] + command.split(), stdin=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)

        try:
            stdout, stderr = process.communicate(sudo_password + '\n', timeout=3)

            if "File exists" in stderr:
                raise FileExistsError(f"The file {file} already exists")

            if process.returncode != 0:
                raise RuntimeError(stderr)
        except subprocess.TimeoutExpired:
            process.kill()
            raise
    else:
        os.unlink(file)


def run_file(file, use_sudo=False, sudo_password=""):
    # Fast fail if there is no file or the file can't be executed
    if file is None or not is_executable(file):
        raise RuntimeError(f"File Execution: File does not exist or have execution permissions [file={file}]")

    if use_sudo:
        command = f"{file}"
        process = subprocess.Popen(['sudo', '-S'] + command.split(), stdin=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)

        try:
            stdout, stderr = process.communicate(sudo_password + '\n', timeout=3)

            if process.returncode != 0:
                raise RuntimeError(stderr)
        except subprocess.TimeoutExpired:
            process.kill()
            raise
    else:
        command_result = subprocess.run(file)

        if command_result.returncode != 0:
            raise RuntimeError(command_result.stderr.decode())


def run_command(command, use_sudo=False, sudo_password="", check_output=True):
    # Fast fail if invalid command is passed in
    if command is None:
        raise RuntimeError(f"Command Execution: Invalid command passed in [command={command}]")

    if use_sudo:
        run_command_with_communication(f"sudo -S {command}", sudo_password, check_output)
    else:
        output = subprocess.run(command, encoding='utf-8', shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        if check_output and output.returncode != 0:
            raise RuntimeError(f"Command Execution: A given command had a non-zero return code [command='{command}', err='{output.stderr}']")

        return output.stdout.strip()


def run_command_with_communication(command, communication, check_output=True):
    process = subprocess.Popen(command.split(), stdin=subprocess.PIPE, stderr=subprocess.PIPE, stdout=subprocess.PIPE, universal_newlines=True)

    try:
        stdout, stderr = process.communicate(communication + '\n')

        if check_output and process.returncode != 0:
            raise RuntimeError(f"Command Execution: A given command had a non-zero return code [command='{command}', err='{stderr}']")

        return stdout.strip()
    except subprocess.TimeoutExpired:
        process.kill()
        raise


"""
Utility functions
"""


def is_moved(origin, destination):
    # Enables fast-failing based on existence
    if not os.path.isfile(destination):
        return False

    if os.path.isfile(origin):
        return False

    return True


def is_broken_link(file):
    return os.path.islink(file) and not os.path.exists(file)


def is_linked(origin, destination):
    return os.path.islink(destination) and os.path.realpath(destination) == os.path.realpath(origin)


def is_copied(origin, destination):
    # Enables fast-failing based on type
    if os.path.islink(destination):
        return False

    # Enables fast-failing based on existence
    if not os.path.isfile(destination):
        return False

    # Enables fast-failing based on file size
    origin_file_size = os.stat(origin).st_size
    destination_file_size = os.stat(destination).st_size

    if origin_file_size != destination_file_size:
        return False

    # Enables fast-failing based on metadata
    origin_last_modified_time = os.path.getmtime(origin)
    destination_last_modified_time = os.path.getmtime(destination)

    if origin_last_modified_time != destination_last_modified_time:
        return False

    # Loads in the files and checks that their hashes are the same
    origin_file_hash = hash_file(origin)
    destination_file_hash = hash_file(destination)
    return origin_file_hash == destination_file_hash


def is_executable(file):
    return os.access(file, os.X_OK)



import platform
from .primitives import MacVersion, VersionRange, Setting


def get_current_mac_version():
    current_version = platform.mac_ver()[0]
    return MacVersion.from_version(current_version)



# General imports
from dataclasses import dataclass
from pathlib import Path


@dataclass(frozen=True, order=True)
class Symlink:
    initial_origin_file: Path
    destination_in_build_package: Path
    script_origin_file: Path
    script_destination_file: Path
    is_sudo: bool


@dataclass(frozen=True, order=True)
class Copy:
    initial_origin_file: Path
    destination_in_build_package: Path
    script_origin_file: Path
    script_destination_file: Path
    is_sudo: bool


@dataclass(frozen=True, order=True)
class RunScript:
    initial_origin_file: Path
    destination_in_build_package: Path
    script_origin_file: Path
    # No need for script destination file since the command is just being executed,
    # and not copied or symlinked elsewhere
    is_sudo: bool



from pathlib import Path

"""
Internal constants
"""
PYDOTFILES_CACHE_DIRECTORY = f"{str(Path.home())}/.pydotfiles"


"""
External defaults
"""
DEFAULT_PYDOTFILES_CONFIG_LOCAL_DIRECTORY = f"{str(Path.home())}/.dotfiles"
DEFAULT_CONFIG_REMOTE_REPO = "https://github.com/JasonYao/pydotfiles-basic.git"



import logging
import shutil
from typing import List
from enum import Enum, auto
from dataclasses import dataclass

from pydotfiles.common import OS, PackageManager
from pydotfiles.utils import run_command, run_command_with_communication


logger = logging.getLogger(__name__)


class LanguagePluginManager(Enum):
    """
    Represents a language plugin manager
    like pyenv-virtualenv
    """
    PYENV_VIRTUALENV = auto()

    @staticmethod
    def from_string(label):
        if label is None:
            raise KeyError("Enum: No language manager passed in")

        return LanguagePluginManager[label.upper().replace("-", "_")]


@dataclass
class VirtualEnvironment:
    version: str
    name: str


class LanguageEnvironmentPluginManager:

    def __init__(self, language_plugin_manager: LanguagePluginManager, virtual_environments: List[VirtualEnvironment]):
        self.language_plugin_manager = language_plugin_manager
        self.virtual_environments = virtual_environments

    @property
    def is_installed(self):
        if self.language_plugin_manager == LanguagePluginManager.PYENV_VIRTUALENV:
            current_installed_brew_packages = run_command("brew list")
            return "pyenv-virtualenv" in current_installed_brew_packages
        else:
            raise NotImplementedError(f"Language Environment Plugin Manager: Sorry, the requested language plugin manager is not currently supported (feel free to open an issue at https://github.com/JasonYao/pydotfiles/issues/new) [env_plugin_manager={self.language_plugin_manager}]")

    def install(self):
        if self.is_installed:
            logger.info(f"Language Environment Plugin Manager: Already installed language environment plugin manager [env_plugin_manager={self.language_plugin_manager}]")
            return

        host_system = OS.get_host_system()
        package_manager = OS.get_package_manager(host_system)

        if package_manager == PackageManager.BREW:
            install_language_environment_plugin_manager_via_homebrew(self.language_plugin_manager)
        else:
            raise NotImplementedError(f"Language Environment Plugin Manager: Sorry, the requested package manager is not currently supported [package_manager={package_manager}]")

    def install_virtual_environments(self):
        for virtual_environment in self.virtual_environments:
            if self.language_plugin_manager == LanguagePluginManager.PYENV_VIRTUALENV:
                install_virtual_environment_version_with_pyenv(virtual_environment)
            else:
                raise NotImplementedError(f"Language Environment Plugin Manager: Sorry, the requested language plugin manager is not currently supported (feel free to open an issue at https://github.com/JasonYao/pydotfiles/issues/new) [env_plugin_manager={self.language_plugin_manager}]")

    def uninstall(self):
        if not self.is_installed:
            logger.info(f"Language Environment Plugin Manager: Already uninstalled language environment plugin manager [env_plugin_manager={self.language_plugin_manager}]")
            return

        host_system = OS.get_host_system()
        package_manager = OS.get_package_manager(host_system)

        if package_manager == PackageManager.BREW:
            uninstall_language_environment_plugin_manager_via_homebrew(self.language_plugin_manager)
        else:
            raise NotImplementedError(f"Language Environment Plugin Manager: Sorry, the requested package manager is not currently supported [package_manager={package_manager}]")

    def uninstall_virtual_environments(self):
        for virtual_environment in self.virtual_environments:
            if self.language_plugin_manager == LanguagePluginManager.PYENV_VIRTUALENV:
                uninstall_virtual_environment_version_with_pyenv(virtual_environment)
            else:
                raise NotImplementedError(f"Language Environment Plugin Manager: Sorry, the requested language plugin manager is not currently supported (feel free to open an issue at https://github.com/JasonYao/pydotfiles/issues/new) [env_plugin_manager={self.language_plugin_manager}]")


"""
Homebrew functions
"""


def install_language_environment_plugin_manager_via_homebrew(language_plugin_manager):
    logger.info(f"Language Environment Plugin Manager: Installing environment manager [env_plugin_manager={language_plugin_manager}]")

    if language_plugin_manager == LanguagePluginManager.PYENV_VIRTUALENV:
        command = "brew install pyenv-virtualenv"
        run_command(command)
    else:
        raise NotImplementedError(f"Language Environment Plugin Manager: Sorry, the requested language manager is not currently supported (feel free to open an issue at https://github.com/JasonYao/pydotfiles/issues/new) [env_plugin_manager={language_plugin_manager}]")

    logger.info(f"Language Environment Plugin Manager: Successfully installed environment manager [env_plugin_manager={language_plugin_manager}]")


def uninstall_language_environment_plugin_manager_via_homebrew(language_plugin_manager):
    logger.info(f"Language Environment Plugin Manager: Uninstalling environment manager [env_plugin_manager={language_plugin_manager}]")

    if language_plugin_manager == LanguagePluginManager.PYENV_VIRTUALENV:
        command = "brew uninstall pyenv-virtualenv"
        run_command(command)
    else:
        raise NotImplementedError(f"Language Environment Plugin Manager: Sorry, the requested language manager is not currently supported (feel free to open an issue at https://github.com/JasonYao/pydotfiles/issues/new) [env_plugin_manager={language_plugin_manager}]")

    logger.info(f"Language Environment Plugin Manager: Successfully uninstalled environment manager [env_plugin_manager={language_plugin_manager}]")


"""
Pyenv functions
"""


def install_virtual_environment_version_with_pyenv(virtual_environment: VirtualEnvironment):
    logger.info(f"Language Environment Plugin Manager: Installing virtual environment [name={virtual_environment.name}, version={virtual_environment.version}]")

    # Makes sure that the virtual env hasn't been created yet
    check_command = "pyenv versions"
    check_response = run_command(check_command)

    if virtual_environment.name in check_response:
        logger.info(f"Language Environment Plugin Manager: Already installed the virtual environment [name={virtual_environment.name}, version={virtual_environment.version}]")
        return

    command = f"pyenv virtualenv {virtual_environment.version} {virtual_environment.name}"
    run_command(command)

    logger.info(f"Language Environment Plugin Manager: Successfully installed virtual environment [name={virtual_environment.name}, version={virtual_environment.version}]")


def uninstall_virtual_environment_version_with_pyenv(virtual_environment: VirtualEnvironment):
    logger.info(f"Language Environment Plugin Manager: Uninstalling virtual environment [name={virtual_environment.name}, version={virtual_environment.version}]")

    # Makes sure that the virtual env hasn't been created yet
    check_command = "pyenv versions"
    check_response = run_command(check_command)

    if virtual_environment.name not in check_response:
        logger.info(f"Language Environment Plugin Manager: Already uninstalled the virtual environment [name={virtual_environment.name}, version={virtual_environment.version}]")
        return

    command = f"pyenv uninstall {virtual_environment.name}"
    run_command_with_communication(command, "y\n")

    logger.info(f"Language Environment Plugin Manager: Successfully uninstalled virtual environment [name={virtual_environment.name}, version={virtual_environment.version}]")



import hashlib
import sys
import logging


"""
General non-project specific helper utilities
"""


class PrettyLogFormatter(logging.Formatter):
    """
    Presentation-layer pretty printing, enabling
    clear guides to an end-user via ANSI escape
    sequences. Should be attached to a logger.

    For more information see:
    https://stackoverflow.com/questions/1343227/can-pythons-logging-format-be-modified-depending-on-the-message-log-level/27835318#27835318
    """

    DEBUG_BLUE = "\033[94m"
    INFO_GREEN = "\033[92m"
    WARNING_YELLOW = "\033[0;33m"
    ERROR_ORANGE = "\033[38:2:255:165:0m"
    FATAL_RED = "\033[31m"

    END_COLOUR = '\033[0m'

    def __init__(self):
        super().__init__()
        self.formatters = {
            logging.DEBUG: logging.Formatter(f"%(asctime)s [{self.DEBUG_BLUE}DEBUG::%(name)s{self.END_COLOUR}]: %(message)s", datefmt=self.date_format),
            logging.INFO: logging.Formatter(f"%(asctime)s [{self.INFO_GREEN}%(levelname)s::%(name)s{self.END_COLOUR}]: %(message)s", datefmt=self.date_format),
            logging.WARNING: logging.Formatter(f"%(asctime)s [{self.WARNING_YELLOW}WARN::%(name)s{self.END_COLOUR}]: %(message)s", datefmt=self.date_format),
            logging.ERROR: logging.Formatter(f"%(asctime)s [{self.ERROR_ORANGE}ERROR::%(name)s{self.END_COLOUR}]: %(message)s", datefmt=self.date_format),
            logging.FATAL: logging.Formatter(f"%(asctime)s [{self.FATAL_RED}FATAL::%(name)s{self.END_COLOUR}]: %(message)s", datefmt=self.date_format),
        }

        self.default_format = logging.Formatter("%(levelname)s: %(name)s: %(message)s")

    @property
    def date_format(self):
        return "%Y-%m-%d %I:%M:%S %p"

    def format(self, record):
        return self.formatters.get(record.levelno, self.default_format).format(record)


class PrettyPrint:
    """
    Presentation-layer pretty printing, enabling
    clear guides to an end-user via ANSI escape
    sequences
    """

    SUCCESS_GREEN = '\033[92m'
    INFO_BLUE = '\033[94m'
    USER_YELLOW = '\033[0;33m'
    WARN_ORANGE = '\033[38:2:255:165:0m'
    FAIL_RED = '\033[31m'

    # Other settings not turned on yet
    # HEADER = '\033[95m'
    # BOLD = '\033[1m'
    # UNDERLINE = '\033[4m'

    END_COLOUR = '\033[0m'

    @staticmethod
    def success(message):
        print(f"[ {PrettyPrint.SUCCESS_GREEN}OK{PrettyPrint.END_COLOUR} ] {message}")

    @staticmethod
    def info(message):
        print(f"[{PrettyPrint.INFO_BLUE}INFO{PrettyPrint.END_COLOUR}] {message}")

    @staticmethod
    def user(message):
        print(f"[{PrettyPrint.USER_YELLOW}USER{PrettyPrint.END_COLOUR}] {message}")

    @staticmethod
    def warn(message):
        print(f"[{PrettyPrint.WARN_ORANGE}WARN{PrettyPrint.END_COLOUR}] {message}")

    @staticmethod
    def fail(message):
        sys.exit(f"[{PrettyPrint.FAIL_RED}FAIL{PrettyPrint.END_COLOUR}] {message}")


class BijectiveDictionary:
    """
    A helpful one-to-one dictionary that has
    some helpful wrappers around a normal dictionary
    with the constraint that the data being input
    is bijective.

    See https://stackoverflow.com/questions/863935/a-data-structure-for-11-mappings-in-python/1374617#1374617
    """

    def __init__(self, input_dictionary=None):
        if input_dictionary is None:
            self.dictionary = {}
        else:
            self.dictionary = input_dictionary.copy()
            for key, value in input_dictionary.items():
                self.dictionary[value] = key

    def add(self, key, value):
        self.dictionary[key] = value
        self.dictionary[value] = key

    def remove(self, key):
        return self.dictionary.pop(self.dictionary.pop(key))

    def get(self, key):
        return self.dictionary.get(key)


"""
General helper functions
"""


def remove_prefix(text, prefix):
    """
    https://stackoverflow.com/questions/16891340/remove-a-prefix-from-a-string/16892491#16892491
    """
    return text[text.startswith(prefix) and len(prefix):]


def hash_file(file_path, block_size=65536):
    """
    Hashes a given file
    """
    hasher = hashlib.sha256()
    with open(file_path, 'rb') as hashed_file:
        buf = hashed_file.read(block_size)
        while len(buf) > 0:
            hasher.update(buf)
            buf = hashed_file.read(block_size)
    return hasher.hexdigest()



import platform
from enum import Enum, auto


class OS(Enum):
    """
    Represents a given operating system
    """
    MACOS = auto()
    LINUX = auto()
    UBUNTU = auto()
    CENTOS = auto()

    @staticmethod
    def get_host_system():
        host_system = platform.system().lower()

        if host_system == "darwin":
            return OS.MACOS
        else:
            raise NotImplementedError(f"OS: The host system's OS is not currently supported [os={host_system}]")

    @staticmethod
    def from_string(label):
        if label is None:
            raise KeyError("Enum: No operating system was passed in")

        if label == "darwin":
            return OS.MACOS

        return OS[label.upper()]

    @staticmethod
    def get_package_manager(os):
        return {
            OS.MACOS: PackageManager.BREW,
            OS.LINUX: PackageManager.APT,
            OS.UBUNTU: PackageManager.APT,
            OS.CENTOS: PackageManager.YUM
        }.get(os)


class PackageManager(Enum):
    """
    Represents a system package manager
    """
    BREW = auto()

    # Linux package managers
    YUM = auto()
    APT = auto()

    @staticmethod
    def from_label(label):
        if label is None:
            raise KeyError("Enum: No package manager was passed in")
        return PackageManager[label.upper()]



# General imports
from dataclasses import dataclass
from enum import Enum, auto
from typing import Dict, Optional, Callable, Any
from pathlib import Path

# Project imports
from .alpha import get_schema as alpha_schema


@dataclass(frozen=True, order=True)
class Configuration:
    file_path: Path
    data: Any


@dataclass(frozen=True, order=True)
class ProfileComposition:
    """
    Indicates how to compose different base profiles together, e.g:
    - work: a composition of core + work profiles
    - personal: a composition of core + dev + personal profiles
    """
    name: str
    profiles: list[str]


class ErrorReason(Enum):
    """
    Represents a particular reason why
    the validation of a given dotfiles,
    module, or file failed
    """

    # Loading errors
    UNSUPPORTED_CONFIG_FILE_TYPE = auto()
    PARSING_ERROR_YAML = auto()

    INVALID_TARGET = auto()

    INVALID_EMPTY_FILE = auto()
    INVALID_SYNTAX = auto()
    INVALID_SCHEMA = auto()

    # Specific schema issues
    INVALID_DATA = auto()
    INVALID_SCHEMA_VERSION = auto()
    INVALID_SCHEMA_TYPE = auto()

    @staticmethod
    def get_help_message(reason):
        help_message_map = {
            ErrorReason.INVALID_EMPTY_FILE: "An empty invalid configuration file was detected",
            ErrorReason.INVALID_SCHEMA: "A given configuration file contains an invalid schema"
        }
        return help_message_map.get(reason)


class PydotfilesError(Exception):

    def __init__(self, reason: ErrorReason, help_message_override: Optional[str] = None, context_map: Optional[Dict] = None):
        super().__init__()
        self.reason = reason
        self.help_message_override = help_message_override
        if context_map is None:
            context_map = {
                "help": "If you need additional context like a stack trace, please run in verbose mode (with the -v flag)"
            }
        self.context_map = context_map

    @property
    def help_message(self):
        undecorated_original_help_message = ErrorReason.get_help_message(self.reason) if self.help_message_override is None else self.help_message_override

        if len(self.context_map) == 0:
            serialized_context_decoration = ""
        else:
            serialized_context_decoration = '\n'.join([f"\t{context_type}: {context_reason}" for context_type, context_reason in self.context_map.items()])

        decorated_error_message = f"{undecorated_original_help_message} [\n{serialized_context_decoration}\n]"
        return decorated_error_message

    @staticmethod
    def due_to_yaml_error() -> 'PydotfilesError':
        return PydotfilesError(ErrorReason.PARSING_ERROR_YAML, "Loader: An invalid YAML syntax error was detected")

    @staticmethod
    def due_to_unsupported_config_filetype() -> 'PydotfilesError':
        return PydotfilesError(ErrorReason.UNSUPPORTED_CONFIG_FILE_TYPE, "Loader: An invalid config filetype could not be parsed (are you using .json or .yaml)?")


def get_configuration_version(version: str, schema: str) -> type:
    configuration_map: Dict[str, Callable] = {
        "alpha": alpha_schema,
    }

    return configuration_map.get(version)(schema)



# General imports
import argparse
import os

# Project imports
from pydotfiles.models import PYDOTFILES_CACHE_DIRECTORY, DEFAULT_PYDOTFILES_CONFIG_LOCAL_DIRECTORY, DEFAULT_CONFIG_REMOTE_REPO
from pydotfiles.models import Dotfiles, CacheDirectory, Validator
from pydotfiles.models import get_pydotfiles_config_data_with_override, load_pydotfiles_config_data, write_pydotfiles_config_data
from pydotfiles.models import PydotfilesError, ValidationError
from pydotfiles.utils import PrettyPrint


class ArgumentDispatcher:
    """
    A presentation-layer class to contain all logic around
    using the pydotfiles API, providing help menus and
    simplifying argument-parsing via dynamic dispatching
    of commands
    """

    def __init__(self, api_arguments):
        self.dotfiles = None
        self.api_arguments = api_arguments

    def download(self, command_arguments):
        help_description = f"""
        Downloads the dotfiles config repo if it hasn't been cloned to local.

        If a remote repo is not provided pydotfiles will use the basic set of pydotfiles
        found at https://github.com/JasonYao/pydotfiles-basic.

        Pydotfiles's configuration is setup in a fallthrough manner:
            - Command-line arguments passed in override all other configs, and will be persisted in $HOME/.pydotfiles/config.json
            - Any non-overridden arguments is then configured from: $HOME/.pydotfiles/config.json (if it exists)
            - Any remaining arguments will default to:
                - Local directory: {DEFAULT_PYDOTFILES_CONFIG_LOCAL_DIRECTORY}
                - Remote repo: {DEFAULT_CONFIG_REMOTE_REPO}
        """
        parser = self.__get_base_parser(help_description, "download")
        parser.add_argument("-l", "--local-directory", help="The local directory where the dotfiles are stored")
        parser.add_argument("-r", "--remote-repo", help="The local directory where the dotfiles are stored")
        args = parser.parse_args(command_arguments)

        # TODO P4: Add in cleaner signature
        config_repo_local, config_repo_remote = get_pydotfiles_config_data_with_override(args.local_directory, args.remote_repo, CacheDirectory())

        self.dotfiles = Dotfiles(config_repo_local, config_repo_remote, args.quiet, args.verbose)

        if self.dotfiles.is_cloned:
            PrettyPrint.success(f"Clone: Dotfiles have already been cloned")
            return

        try:
            self.dotfiles.download()
        except PydotfilesError as e:
            PrettyPrint.fail(e.help_message)

    def install(self, command_arguments):
        help_description = """
        Installs your dotfile's modules (default: installs all modules)
        NOTE: Your dotfiles need to have first been downloaded via `pydotfiles download` beforehand
        """
        parser = self.__get_base_parser(help_description, "install")
        parser.add_argument("-m", "--modules", help="A list of specific modules to install", nargs="+")
        args = parser.parse_args(command_arguments)

        # TODO P4: Add in cleaner signature
        config_repo_local, config_repo_remote = load_pydotfiles_config_data(CacheDirectory())

        self.dotfiles = Dotfiles(config_repo_local, config_repo_remote, args.quiet, args.verbose, args.modules)

        if not self.dotfiles.is_cloned:
            PrettyPrint.fail(f"Install: No dotfiles detected, please download it first with `pydotfiles download`")

        if args.modules is None:
            self.dotfiles.install_all()
        else:
            self.dotfiles.install_multiple_modules(args.modules)

    def uninstall(self, command_arguments):
        help_description = """
        Uninstalls your dotfile's modules (default: uninstalls all modules, but leaves packages, applications, and dev-environments alone)
        """
        parser = self.__get_base_parser(help_description, "uninstall")
        parser.add_argument("-m", "--modules", help="A list of specific modules to uninstall", nargs="+")
        parser.add_argument("-p", "--uninstall-packages", help="Will uninstall all packages installed with these module(s)", action="store_true")
        parser.add_argument("-a", "--uninstall-applications", help="Will uninstall all applications installed with these module(s)", action="store_true")
        parser.add_argument("-e", "--uninstall-environments", help="Will uninstall all dev environments with these module(s)", action="store_true")
        args = parser.parse_args(command_arguments)

        config_repo_local, config_repo_remote = load_pydotfiles_config_data(CacheDirectory())

        PrettyPrint.info(f"Uninstall: Uninstalling dotfiles")

        self.dotfiles = Dotfiles(config_repo_local, config_repo_remote, args.quiet, args.verbose, args.modules)

        if not self.dotfiles.is_cloned:
            PrettyPrint.fail(f"Uninstall: Could not uninstall- no dotfiles detected")

        if args.modules is None:
            self.dotfiles.uninstall_all(args.uninstall_packages, args.uninstall_applications, args.uninstall_environments)
        else:
            self.dotfiles.uninstall_multiple_modules(args.modules, args.uninstall_packages, args.uninstall_applications, args.uninstall_environments)

    def update(self, command_arguments):
        help_description = """
        Updates the local dotfiles from the remote repo
        """
        parser = self.__get_base_parser(help_description, "update")
        args = parser.parse_args(command_arguments)

        config_repo_local, config_repo_remote = load_pydotfiles_config_data(CacheDirectory())

        self.dotfiles = Dotfiles(config_repo_local, config_repo_remote, args.quiet, args.verbose)

        if not self.dotfiles.is_cloned:
            PrettyPrint.fail(f"Update: Could not update- no dotfiles detected")

        self.dotfiles.update()

    def clean(self, command_arguments):
        help_description = f"""
        Deletes either the pydotfiles cache or the downloaded local dotfiles config repo

        Possible choices:
            - cache: Deletes everything in the pydotfiles cache directory ({os.path.expanduser(PYDOTFILES_CACHE_DIRECTORY)})
            - repo: Deletes everything in the locally downloaded dotfiles configuration directory ({DEFAULT_PYDOTFILES_CONFIG_LOCAL_DIRECTORY})
        """
        valid_cleaning_targets = ['cache', 'repo']
        parser = self.__get_base_parser(help_description, "clean")
        parser.add_argument('clean_target', help='Clears out the given cleaning target', choices=valid_cleaning_targets)
        args = parser.parse_args(command_arguments)

        config_repo_local, config_repo_remote = load_pydotfiles_config_data(CacheDirectory())

        self.dotfiles = Dotfiles(config_repo_local, config_repo_remote, args.quiet, args.verbose)

        self.dotfiles.clean(args.clean_target)

    def set(self, command_arguments):
        help_description = f"""
        Enables direct setting of pydotfile config values
        """
        parser = self.__get_base_parser(help_description, "set")
        parser.add_argument("-l", "--local-directory", help="Sets the dotfiles configuration repo to a different local directory")
        parser.add_argument("-r", "--remote-repo", help="Sets pydotfiles to point to a different remote repo")
        args = parser.parse_args(command_arguments)

        cache_directory = CacheDirectory()

        config_repo_local, config_repo_remote = load_pydotfiles_config_data(cache_directory)

        if args.local_directory is not None:
            config_repo_local = args.local_directory

        if args.remote_repo is not None:
            config_repo_remote = args.remote_repo

        write_pydotfiles_config_data(cache_directory, config_repo_local, config_repo_remote)
        PrettyPrint.success(f"Set: Successfully persisted configuration data [local-directory={config_repo_local}, remote-repo={config_repo_remote}]")

    def validate(self, command_arguments):
        help_description = """
        Validates a given directory and whether it's pydotfiles-compliant.
        (default: Checks the current working directory)
        """
        parser = self.__get_base_parser(help_description, "validate")
        parser.add_argument("-d", "--directory", help="Validates the passed in directory", default=os.getcwd())

        args = parser.parse_args(command_arguments)

        validator = Validator(args.quiet, args.verbose)
        try:
            validator.validate_directory(args.directory)
        except ValidationError as e:
            PrettyPrint.fail(e.help_message)

    @staticmethod
    def __get_base_parser(description, sub_command):
        parser = argparse.ArgumentParser(
            prog=f"pydotfiles {sub_command}",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=description
        )
        logging_parser_group = parser.add_mutually_exclusive_group()
        logging_parser_group.add_argument("-v", "--verbose", help="Enables more verbose logging", action="store_true")
        logging_parser_group.add_argument("-q", "--quiet", help="Squelches the default logging (still outputs to stderr upon failures)", action="store_true")
        return parser



import plistlib
from typing import Dict, List
from pydotfiles.models.utils import load_data_from_file
from pydotfiles.defaults import MacVersion, VersionRange, Setting

from pydotfiles.environments import VirtualEnvironment
from pydotfiles.environments import LanguagePluginManager, LanguageEnvironmentPluginManager
from pydotfiles.environments import LanguageManager, LanguageEnvironmentManager
from pydotfiles.environments import DevelopmentEnvironment


def get_os_default_settings(default_setting_file_path):
    # Loads in the default settings file
    default_settings_data = load_data_from_file(default_setting_file_path)

    # Parses the default settings data
    return parse_default_settings(default_settings_data)


"""
Loading methods
"""


def load_plist(plist_path):
    with open(plist_path, 'rb') as plist_file:
        return plistlib.load(plist_file)


"""
Parsing methods
"""


def parse_default_settings(default_settings_data):
    if not default_settings_data:
        default_settings_data = []

    version = default_settings_data.get("version")
    schema_type = default_settings_data.get("schema")

    if version != "alpha":
        raise NotImplementedError(f"Loading: Unable to load default settings file with an unsupported version number [found_version={version}]")

    if schema_type != "default_settings":
        raise ValueError(f"Loading: Invalid data file was passed in based on detected schema type [schema_type={schema_type}]")

    return alpha_default_parse_data(default_settings_data.get("default_settings"))


def parse_developer_environments(developer_environments_data):
    if developer_environments_data is None:
        return []

    return alpha_developer_environments_parse_data(developer_environments_data)


"""
Version-based parsers
"""


def alpha_default_parse_data(default_settings_data):
    settings = []
    for raw_setting in default_settings_data:
        name = raw_setting.get("name")
        enabled = raw_setting.get("enabled", True)
        description = raw_setting.get("description")

        raw_start = raw_setting.get("start")
        start = None if raw_start is None else MacVersion.from_name(raw_start)

        raw_end = raw_setting.get("end")
        end = None if raw_end is None else MacVersion.from_name(raw_end)

        valid_version_range = VersionRange(start, end)
        command = raw_setting.get("command")
        check_command = raw_setting.get("check_command")
        expected_check_state = raw_setting.get("expected_check_state")

        run_as_sudo = raw_setting.get("sudo", False)

        check_output = raw_setting.get("check_output", True)
        settings.append(Setting(
            name=name,
            valid_version_range=valid_version_range,
            command=command,
            enabled=enabled,
            description=description,
            check_command=check_command,
            expected_check_state=expected_check_state,
            run_as_sudo=run_as_sudo,
            check_output=check_output,
        ))

    return settings


def alpha_developer_environments_parse_data(developer_environments_data):
    developer_environments = []

    for raw_developer_environment in developer_environments_data:
        language = raw_developer_environment.get("language")
        versions = raw_developer_environment.get("versions")
        language_environment_manager = parse_language_environment_manager(raw_developer_environment.get("environment_manager"))

        developer_environments.append(DevelopmentEnvironment(
            language=language,
            versions=versions,
            language_environment_manager=language_environment_manager
        ))
    return developer_environments


def parse_language_environment_manager(environment_manager_data: Dict):
    environment_manager = LanguageManager.from_string(environment_manager_data.get("name"))
    plugin_managers = parse_language_environment_plugins(environment_manager_data.get("plugins"))

    return LanguageEnvironmentManager(environment_manager, language_plugin_managers=plugin_managers)


def parse_language_environment_plugins(plugins_data: List[Dict]):
    if plugins_data is None:
        return []

    plugin_managers = []
    for plugin_data in plugins_data:
        language_plugin = LanguagePluginManager.from_string(plugin_data.get("name"))
        virtual_environments_data = parse_virtual_environments(plugin_data.get("virtual_environments"))
        plugin_managers.append(LanguageEnvironmentPluginManager(language_plugin, virtual_environments_data))
    return plugin_managers


def parse_virtual_environments(virtual_environments_data: List[Dict]):
    if virtual_environments_data is None:
        return []
    return [VirtualEnvironment(virtual_environment.get("version"), virtual_environment.get("name")) for virtual_environment in virtual_environments_data]
